from .cars3d import Cars
from .generic import GenericDataset
from .mnist import NoisyMNIST, SplitMNIST, TangledMNIST, MNISTSVHN
from .wiw import WIW
from .xrmb import XRMB
from .mfeat import MFeat
from .twitter import Twitter
from .cub import Cub
